@@ -1,30 +1,8 @@
 /*
- * Copyright (c) 1999-2008 Douglas Gilbert.
+ * Copyright (c) 1999-2011 Douglas Gilbert.
  * All rights reserved.
- *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions
- * are met:
- * 1. Redistributions of source code must retain the above copyright
- *    notice, this list of conditions and the following disclaimer.
- * 2. Redistributions in binary form must reproduce the above copyright
- *    notice, this list of conditions and the following disclaimer in the
- *    documentation and/or other materials provided with the distribution.
- * 3. The name of the author may not be used to endorse or promote products
- *    derived from this software without specific prior written permission.
- *
- * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
- * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
- * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
- * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
- * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
- * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
- * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
- * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
- * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
- * SUCH DAMAGE.
- *
+ * Use of this source code is governed by a BSD-style
+ * license that can be found in the BSD_LICENSE file.
  */
 
 #include <stdio.h>
@@ -39,6 +17,9 @@
 #include "sg_cmds_extra.h"
 #include "sg_pt.h"
 
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
 
 
 #define SENSE_BUFF_LEN 32       /* Arbitrary, could be larger */
@@ -65,6 +46,8 @@
 #define PERSISTENT_RESERVE_IN_CMDLEN 10
 #define PERSISTENT_RESERVE_OUT_CMD 0x5f
 #define PERSISTENT_RESERVE_OUT_CMDLEN 10
+#define READ_BLOCK_LIMITS_CMD 0x5
+#define READ_BLOCK_LIMITS_CMDLEN 6
 #define READ_BUFFER_CMD 0x3c
 #define READ_BUFFER_CMDLEN 10
 #define READ_DEFECT10_CMD     0x37
@@ -79,13 +62,18 @@
 #define SERVICE_ACTION_IN_12_CMDLEN 12
 #define READ_LONG10_CMD 0x3e
 #define READ_LONG10_CMDLEN 10
+#define UNMAP_CMD 0x42
+#define UNMAP_CMDLEN 10
+#define VERIFY10_CMD 0x2f
+#define VERIFY10_CMDLEN 10
+#define VERIFY16_CMD 0x8f
+#define VERIFY16_CMDLEN 16
 #define WRITE_LONG10_CMD 0x3f
 #define WRITE_LONG10_CMDLEN 10
 #define WRITE_BUFFER_CMD 0x3b
 #define WRITE_BUFFER_CMDLEN 10
-#define VERIFY10_CMD 0x2f
-#define VERIFY10_CMDLEN 10
 
+#define GET_LBA_STATUS_SA 0x12
 #define READ_LONG_16_SA 0x11
 #define READ_MEDIA_SERIAL_NUM_SA 0x1
 #define REPORT_IDENTIFYING_INFORMATION_SA 0x5
@@ -93,8 +81,89 @@
 #define SET_IDENTIFYING_INFORMATION_SA 0x6
 #define SET_TGT_PRT_GRP_SA 0xa
 #define WRITE_LONG_16_SA 0x11
+#define REPORT_REFERRALS_SA 0x13
 
 
+/* Invokes a SCSI GET LBA STATUS command (SBC). Returns 0 -> success,
+ * SG_LIB_CAT_INVALID_OP -> GET LBA STATUS not supported,
+ * SG_LIB_CAT_ILLEGAL_REQ -> bad field in cdb, SG_LIB_CAT_ABORTED_COMMAND,
+ * SG_LIB_CAT_NOT_READY -> device not ready, -1 -> other failure */
+int
+sg_ll_get_lba_status(int sg_fd, uint64_t start_llba, void * resp,
+                     int alloc_len, int noisy, int verbose)
+{
+    int k, res, sense_cat, ret;
+    unsigned char getLbaStatCmd[SERVICE_ACTION_IN_16_CMDLEN];
+    unsigned char sense_b[SENSE_BUFF_LEN];
+    struct sg_pt_base * ptvp;
+
+    memset(getLbaStatCmd, 0, sizeof(getLbaStatCmd));
+    getLbaStatCmd[0] = SERVICE_ACTION_IN_16_CMD;
+    getLbaStatCmd[1] = GET_LBA_STATUS_SA;
+
+    getLbaStatCmd[2] = (start_llba >> 56) & 0xff;
+    getLbaStatCmd[3] = (start_llba >> 48) & 0xff;
+    getLbaStatCmd[4] = (start_llba >> 40) & 0xff;
+    getLbaStatCmd[5] = (start_llba >> 32) & 0xff;
+    getLbaStatCmd[6] = (start_llba >> 24) & 0xff;
+    getLbaStatCmd[7] = (start_llba >> 16) & 0xff;
+    getLbaStatCmd[8] = (start_llba >> 8) & 0xff;
+    getLbaStatCmd[9] = start_llba & 0xff;
+    getLbaStatCmd[10] = (alloc_len >> 24) & 0xff;
+    getLbaStatCmd[11] = (alloc_len >> 16) & 0xff;
+    getLbaStatCmd[12] = (alloc_len >> 8) & 0xff;
+    getLbaStatCmd[13] = alloc_len & 0xff;
+    if (NULL == sg_warnings_strm)
+        sg_warnings_strm = stderr;
+    if (verbose) {
+        fprintf(sg_warnings_strm, "    Get LBA status cmd: ");
+        for (k = 0; k < SERVICE_ACTION_IN_16_CMDLEN; ++k)
+            fprintf(sg_warnings_strm, "%02x ", getLbaStatCmd[k]);
+        fprintf(sg_warnings_strm, "\n");
+    }
+
+    ptvp = construct_scsi_pt_obj();
+    if (NULL == ptvp) {
+        fprintf(sg_warnings_strm, "get LBA status: out of memory\n");
+        return -1;
+    }
+    set_scsi_pt_cdb(ptvp, getLbaStatCmd, sizeof(getLbaStatCmd));
+    set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
+    set_scsi_pt_data_in(ptvp, (unsigned char *)resp, alloc_len);
+    res = do_scsi_pt(ptvp, sg_fd, DEF_PT_TIMEOUT, verbose);
+    ret = sg_cmds_process_resp(ptvp, "get LBA status", res, alloc_len,
+                               sense_b, noisy, verbose, &sense_cat);
+    if (-1 == ret)
+        ;
+    else if (-2 == ret) {
+        switch (sense_cat) {
+        case SG_LIB_CAT_NOT_READY:
+        case SG_LIB_CAT_ILLEGAL_REQ:
+        case SG_LIB_CAT_INVALID_OP:
+        case SG_LIB_CAT_UNIT_ATTENTION:
+        case SG_LIB_CAT_ABORTED_COMMAND:
+            ret = sense_cat;
+            break;
+        case SG_LIB_CAT_RECOVERED:
+        case SG_LIB_CAT_NO_SENSE:
+            ret = 0;
+            break;
+        default:
+            ret = -1;
+            break;
+        }
+    } else {
+        if ((verbose > 2) && (ret > 0)) {
+            fprintf(sg_warnings_strm, "    get LBA status: response%s\n",
+                    (ret > 256 ? ", first 256 bytes" : ""));
+            dStrHex((const char *)resp, (ret > 256 ? 256 : ret), -1);
+        }
+        ret = 0;
+    }
+    destruct_scsi_pt_obj(ptvp);
+    return ret;
+}
+
 /* Invokes a SCSI REPORT TARGET PORT GROUPS command. Return of 0 -> success,
  * SG_LIB_CAT_INVALID_OP -> Report Target Port Groups not supported,
  * SG_LIB_CAT_ILLEGAL_REQ -> bad field in cdb, SG_LIB_CAT_ABORTED_COMMAND,
@@ -154,8 +223,14 @@ sg_ll_report_tgt_prt_grp(int sg_fd, void * resp, int mx_resp_len, int noisy,
             ret = -1;
             break;
         }
-    } else
+    } else {
+        if ((verbose > 2) && (ret > 0)) {
+            fprintf(sg_warnings_strm, "    report target port group: "
+                    "response%s\n", (ret > 256 ? ", first 256 bytes" : ""));
+            dStrHex((const char *)resp, (ret > 256 ? 256 : ret), -1);
+        }
         ret = 0;
+    }
     destruct_scsi_pt_obj(ptvp);
     return ret;
 }
@@ -229,6 +304,86 @@ sg_ll_set_tgt_prt_grp(int sg_fd, void * paramp, int param_len, int noisy,
     return ret;
 }
 
+/* Invokes a SCSI REPORT REFERRALS command. Return of 0 -> success,
+ * SG_LIB_CAT_INVALID_OP -> Report Referrals not supported,
+ * SG_LIB_CAT_ILLEGAL_REQ -> bad field in cdb, SG_LIB_CAT_ABORTED_COMMAND,
+ * SG_LIB_CAT_UNIT_ATTENTION, -1 -> other failure */
+int
+sg_ll_report_referrals(int sg_fd, uint64_t start_llba, int one_seg, void * resp,
+                       int mx_resp_len, int noisy, int verbose)
+{
+    int k, res, ret, sense_cat;
+    unsigned char repRefCmdBlk[SERVICE_ACTION_IN_16_CMDLEN] =
+                         {SERVICE_ACTION_IN_16_CMD, REPORT_REFERRALS_SA,
+                          0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
+    unsigned char sense_b[SENSE_BUFF_LEN];
+    struct sg_pt_base * ptvp;
+
+    repRefCmdBlk[2] = (start_llba >> 56) & 0xff;
+    repRefCmdBlk[3] = (start_llba >> 48) & 0xff;
+    repRefCmdBlk[4] = (start_llba >> 40) & 0xff;
+    repRefCmdBlk[5] = (start_llba >> 32) & 0xff;
+    repRefCmdBlk[6] = (start_llba >> 24) & 0xff;
+    repRefCmdBlk[7] = (start_llba >> 16) & 0xff;
+    repRefCmdBlk[8] = (start_llba >> 8) & 0xff;
+    repRefCmdBlk[9] = start_llba & 0xff;
+    repRefCmdBlk[10] = (mx_resp_len >> 24) & 0xff;
+    repRefCmdBlk[11] = (mx_resp_len >> 16) & 0xff;
+    repRefCmdBlk[12] = (mx_resp_len >> 8) & 0xff;
+    repRefCmdBlk[13] = mx_resp_len & 0xff;
+    repRefCmdBlk[14] = one_seg & 0x1;
+    if (NULL == sg_warnings_strm)
+        sg_warnings_strm = stderr;
+    if (verbose) {
+        fprintf(sg_warnings_strm, "    report referrals cdb: ");
+        for (k = 0; k < SERVICE_ACTION_IN_16_CMDLEN; ++k)
+            fprintf(sg_warnings_strm, "%02x ", repRefCmdBlk[k]);
+        fprintf(sg_warnings_strm, "\n");
+    }
+
+    ptvp = construct_scsi_pt_obj();
+    if (NULL == ptvp) {
+        fprintf(sg_warnings_strm, "report target port groups: out of "
+                "memory\n");
+        return -1;
+    }
+    set_scsi_pt_cdb(ptvp, repRefCmdBlk, sizeof(repRefCmdBlk));
+    set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
+    set_scsi_pt_data_in(ptvp, (unsigned char *)resp, mx_resp_len);
+    res = do_scsi_pt(ptvp, sg_fd, DEF_PT_TIMEOUT, verbose);
+    ret = sg_cmds_process_resp(ptvp, "report referrals", res,
+                               mx_resp_len, sense_b, noisy, verbose,
+                               &sense_cat);
+    if (-1 == ret)
+        ;
+    else if (-2 == ret) {
+        switch (sense_cat) {
+        case SG_LIB_CAT_INVALID_OP:
+        case SG_LIB_CAT_ILLEGAL_REQ:
+        case SG_LIB_CAT_UNIT_ATTENTION:
+        case SG_LIB_CAT_ABORTED_COMMAND:
+            ret = sense_cat;
+            break;
+        case SG_LIB_CAT_RECOVERED:
+        case SG_LIB_CAT_NO_SENSE:
+            ret = 0;
+            break;
+        default:
+            ret = -1;
+            break;
+        }
+    } else {
+        if ((verbose > 2) && (ret > 0)) {
+            fprintf(sg_warnings_strm, "    report referrals: response%s\n",
+                    (ret > 256 ? ", first 256 bytes" : ""));
+            dStrHex((const char *)resp, (ret > 256 ? 256 : ret), -1);
+        }
+        ret = 0;
+    }
+    destruct_scsi_pt_obj(ptvp);
+    return ret;
+}
+
 /* Invokes a SCSI SEND DIAGNOSTIC command. Foreground, extended self tests can
  * take a long time, if so set long_duration flag. Return of 0 -> success,
  * SG_LIB_CAT_INVALID_OP -> Send diagnostic not supported,
@@ -365,8 +520,14 @@ sg_ll_receive_diag(int sg_fd, int pcv, int pg_code, void * resp,
             ret = -1;
             break;
         }
-    } else
+    } else {
+        if ((verbose > 2) && (ret > 0)) {
+            fprintf(sg_warnings_strm, "    receive diagnostic results: "
+                    "response%s\n", (ret > 256 ? ", first 256 bytes" : ""));
+            dStrHex((const char *)resp, (ret > 256 ? 256 : ret), -1);
+        }
         ret = 0;
+    }
     destruct_scsi_pt_obj(ptvp);
     return ret;
 }
@@ -1056,8 +1217,14 @@ sg_ll_read_long10(int sg_fd, int pblock, int correct, unsigned int lba,
             ret = -1;
             break;
         }
-    } else
+    } else {
+        if ((verbose > 2) && (ret > 0)) {
+            fprintf(sg_warnings_strm, "    read long(10): response%s\n",
+                    (ret > 256 ? ", first 256 bytes" : ""));
+            dStrHex((const char *)resp, (ret > 256 ? 256 : ret), -1);
+        }
         ret = 0;
+    }
     destruct_scsi_pt_obj(ptvp);
     return ret;
 }
@@ -1155,8 +1322,14 @@ sg_ll_read_long16(int sg_fd, int pblock, int correct, uint64_t llba,
             ret = -1;
             break;
         }
-    } else
+    } else {
+        if ((verbose > 2) && (ret > 0)) {
+            fprintf(sg_warnings_strm, "    read long(16): response%s\n",
+                    (ret > 256 ? ", first 256 bytes" : ""));
+            dStrHex((const char *)resp, (ret > 256 ? 256 : ret), -1);
+        }
         ret = 0;
+    }
     destruct_scsi_pt_obj(ptvp);
     return ret;
 }
@@ -1452,6 +1625,102 @@ sg_ll_verify10(int sg_fd, int vrprotect, int dpo, int bytechk,
     return ret;
 }
 
+/* Invokes a SCSI VERIFY (16) command (SBC and MMC).
+ * Note that 'veri_len' is in blocks while 'data_out_len' is in bytes.
+ * Returns of 0 -> success,
+ * SG_LIB_CAT_INVALID_OP -> Verify(16) not supported,
+ * SG_LIB_CAT_ILLEGAL_REQ -> bad field in cdb, SG_LIB_CAT_UNIT_ATTENTION,
+ * SG_LIB_CAT_MEDIUM_HARD -> medium or hardware error, no valid info,
+ * SG_LIB_CAT_MEDIUM_HARD_WITH_INFO -> as previous, with valid info,
+ * SG_LIB_CAT_NOT_READY -> device not ready, SG_LIB_CAT_ABORTED_COMMAND,
+ * -1 -> other failure */
+int
+sg_ll_verify16(int sg_fd, int vrprotect, int dpo, int bytechk, uint64_t llba,
+               int veri_len, int group_num, void * data_out,
+               int data_out_len, uint64_t * infop, int noisy, int verbose)
+{
+    int k, res, ret, sense_cat;
+    unsigned char vCmdBlk[VERIFY16_CMDLEN] =
+                {VERIFY16_CMD, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
+    unsigned char sense_b[SENSE_BUFF_LEN];
+    struct sg_pt_base * ptvp;
+
+    vCmdBlk[1] = ((vrprotect & 0x7) << 5) | ((dpo & 0x1) << 4) |
+                 ((bytechk & 0x1) << 1) ;
+    vCmdBlk[2] = (llba >> 56) & 0xff;
+    vCmdBlk[3] = (llba >> 48) & 0xff;
+    vCmdBlk[4] = (llba >> 40) & 0xff;
+    vCmdBlk[5] = (llba >> 32) & 0xff;
+    vCmdBlk[6] = (llba >> 24) & 0xff;
+    vCmdBlk[7] = (llba >> 16) & 0xff;
+    vCmdBlk[8] = (llba >> 8) & 0xff;
+    vCmdBlk[9] = llba & 0xff;
+    vCmdBlk[10] = (veri_len >> 24) & 0xff;
+    vCmdBlk[11] = (veri_len >> 16) & 0xff;
+    vCmdBlk[12] = (veri_len >> 8) & 0xff;
+    vCmdBlk[13] = veri_len & 0xff;
+    vCmdBlk[14] = group_num & 0x1f;
+    if (NULL == sg_warnings_strm)
+        sg_warnings_strm = stderr;
+    if (verbose > 1) {
+        fprintf(sg_warnings_strm, "    Verify(16) cdb: ");
+        for (k = 0; k < VERIFY16_CMDLEN; ++k)
+            fprintf(sg_warnings_strm, "%02x ", vCmdBlk[k]);
+        fprintf(sg_warnings_strm, "\n");
+    }
+    ptvp = construct_scsi_pt_obj();
+    if (NULL == ptvp) {
+        fprintf(sg_warnings_strm, "verify (16): out of memory\n");
+        return -1;
+    }
+    set_scsi_pt_cdb(ptvp, vCmdBlk, sizeof(vCmdBlk));
+    set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
+    if (data_out_len > 0)
+        set_scsi_pt_data_out(ptvp, (unsigned char *)data_out, data_out_len);
+    res = do_scsi_pt(ptvp, sg_fd, DEF_PT_TIMEOUT, verbose);
+    ret = sg_cmds_process_resp(ptvp, "verify (16)", res, 0, sense_b,
+                               noisy, verbose, &sense_cat);
+    if (-1 == ret)
+        ;
+    else if (-2 == ret) {
+        switch (sense_cat) {
+        case SG_LIB_CAT_NOT_READY:
+        case SG_LIB_CAT_INVALID_OP:
+        case SG_LIB_CAT_ILLEGAL_REQ:
+        case SG_LIB_CAT_UNIT_ATTENTION:
+        case SG_LIB_CAT_ABORTED_COMMAND:
+            ret = sense_cat;
+            break;
+        case SG_LIB_CAT_RECOVERED:
+        case SG_LIB_CAT_NO_SENSE:
+            ret = 0;
+            break;
+        case SG_LIB_CAT_MEDIUM_HARD:
+            {
+                int valid, slen;
+                uint64_t ull = 0;
+
+                slen = get_scsi_pt_sense_len(ptvp);
+                valid = sg_get_sense_info_fld(sense_b, slen, &ull);
+                if (valid) {
+                    if (infop)
+                        *infop = ull;
+                    ret = SG_LIB_CAT_MEDIUM_HARD_WITH_INFO;
+                } else
+                    ret = SG_LIB_CAT_MEDIUM_HARD;
+            }
+            break;
+        default:
+            ret = -1;
+            break;
+        }
+    } else
+        ret = 0;
+
+    destruct_scsi_pt_obj(ptvp);
+    return ret;
+}
+
 /* Invokes a ATA PASS-THROUGH (12 or 16) SCSI command (SAT). If cdb_len
  * is 12 then a ATA PASS-THROUGH (12) command is called. If cdb_len is 16
  * then a ATA PASS-THROUGH (16) command is called. If cdb_len is any other
@@ -1760,3 +2029,146 @@ sg_ll_write_buffer(int sg_fd, int mode, int buffer_id, int buffer_offset,
     destruct_scsi_pt_obj(ptvp);
     return ret;
 }
+
+/* Invokes a SCSI UNMAP command. Return of 0 -> success,
+ * SG_LIB_CAT_INVALID_OP -> command not supported,
+ * SG_LIB_CAT_ILLEGAL_REQ -> bad field in cdb, SG_LIB_CAT_ABORTED_COMMAND,
+ * SG_LIB_CAT_UNIT_ATTENTION, -1 -> other failure */
+int
+sg_ll_unmap(int sg_fd, int group_num, int timeout_secs, void * paramp,
+            int param_len, int noisy, int verbose)
+{
+    return sg_ll_unmap_v2(sg_fd, 0, group_num, timeout_secs, paramp,
+                          param_len, noisy, verbose);
+}
+
+/* Invokes a SCSI UNMAP (SBC-3) command. Version 2 adds anchor field
+ * (sbc3r22). Otherwise same as sg_ll_unmap() . */
+int
+sg_ll_unmap_v2(int sg_fd, int anchor, int group_num, int timeout_secs,
+               void * paramp, int param_len, int noisy, int verbose)
+{
+    int k, res, ret, sense_cat, tmout;
+    unsigned char uCmdBlk[UNMAP_CMDLEN] =
+                         {UNMAP_CMD, 0, 0, 0, 0, 0, 0, 0, 0, 0};
+    unsigned char sense_b[SENSE_BUFF_LEN];
+    struct sg_pt_base * ptvp;
+
+    if (anchor)
+        uCmdBlk[1] |= 0x1;
+    tmout = (timeout_secs > 0) ? timeout_secs : DEF_PT_TIMEOUT;
+    uCmdBlk[7] = group_num & 0x1f;
+    uCmdBlk[7] = (param_len >> 8) & 0xff;
+    uCmdBlk[8] = param_len & 0xff;
+    if (NULL == sg_warnings_strm)
+        sg_warnings_strm = stderr;
+    if (verbose) {
+        fprintf(sg_warnings_strm, "    unmap cdb: ");
+        for (k = 0; k < UNMAP_CMDLEN; ++k)
+            fprintf(sg_warnings_strm, "%02x ", uCmdBlk[k]);
+        fprintf(sg_warnings_strm, "\n");
+        if ((verbose > 1) && paramp && param_len) {
+            fprintf(sg_warnings_strm, "    unmap parameter list:\n");
+            dStrHex((const char *)paramp, param_len, -1);
+        }
+    }
+
+    ptvp = construct_scsi_pt_obj();
+    if (NULL == ptvp) {
+        fprintf(sg_warnings_strm, "unmap: out of memory\n");
+        return -1;
+    }
+    set_scsi_pt_cdb(ptvp, uCmdBlk, sizeof(uCmdBlk));
+    set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
+    set_scsi_pt_data_out(ptvp, (unsigned char *)paramp, param_len);
+    res = do_scsi_pt(ptvp, sg_fd, tmout, verbose);
+    ret = sg_cmds_process_resp(ptvp, "unmap", res, 0,
+                               sense_b, noisy, verbose, &sense_cat);
+    if (-1 == ret)
+        ;
+    else if (-2 == ret) {
+        switch (sense_cat) {
+        case SG_LIB_CAT_INVALID_OP:
+        case SG_LIB_CAT_ILLEGAL_REQ:
+        case SG_LIB_CAT_UNIT_ATTENTION:
+        case SG_LIB_CAT_ABORTED_COMMAND:
+            ret = sense_cat;
+            break;
+        case SG_LIB_CAT_RECOVERED:
+        case SG_LIB_CAT_NO_SENSE:
+            ret = 0;
+            break;
+        default:
+            ret = -1;
+            break;
+        }
+    } else
+        ret = 0;
+    destruct_scsi_pt_obj(ptvp);
+    return ret;
+}
+
+/* Invokes a SCSI READ BLOCK LIMITS command. Return of 0 -> success,
+ * SG_LIB_CAT_INVALID_OP -> Read block limits not supported,
+ * SG_LIB_CAT_ILLEGAL_REQ -> bad field in cdb,
+ * SG_LIB_CAT_ABORTED_COMMAND,
+ * SG_LIB_NOT_READY (shouldn't happen), -1 -> other failure */
+int
+sg_ll_read_block_limits(int sg_fd, void * resp, int mx_resp_len,
+                        int noisy, int verbose)
+{
+    int k, ret, res, sense_cat;
+    unsigned char rlCmdBlk[READ_BLOCK_LIMITS_CMDLEN] = 
+      {READ_BLOCK_LIMITS_CMD, 0, 0, 0, 0, 0};
+    unsigned char sense_b[SENSE_BUFF_LEN];
+    struct sg_pt_base * ptvp;
+
+    if (NULL == sg_warnings_strm)
+        sg_warnings_strm = stderr;
+    if (verbose) {
+        fprintf(sg_warnings_strm, "    read block limits cdb: ");
+        for (k = 0; k < READ_BLOCK_LIMITS_CMDLEN; ++k)
+            fprintf(sg_warnings_strm, "%02x ", rlCmdBlk[k]);
+        fprintf(sg_warnings_strm, "\n");
+    }
+
+    ptvp = construct_scsi_pt_obj();
+    if (NULL == ptvp) {
+        fprintf(sg_warnings_strm, "read block limits: out of memory\n");
+        return -1;
+    }
+    set_scsi_pt_cdb(ptvp, rlCmdBlk, sizeof(rlCmdBlk));
+    set_scsi_pt_sense(ptvp, sense_b, sizeof(sense_b));
+    set_scsi_pt_data_in(ptvp, (unsigned char *)resp, mx_resp_len);
+    res = do_scsi_pt(ptvp, sg_fd, DEF_PT_TIMEOUT, verbose);
+    ret = sg_cmds_process_resp(ptvp, "read block limits", res, mx_resp_len,
+                               sense_b, noisy, verbose, &sense_cat);
+    if (-1 == ret)
+        ;
+    else if (-2 == ret) {
+        switch (sense_cat) {
+        case SG_LIB_CAT_INVALID_OP:
+        case SG_LIB_CAT_ILLEGAL_REQ:
+        case SG_LIB_CAT_ABORTED_COMMAND:
+        case SG_LIB_CAT_NOT_READY:      /* shouldn't happen ?? */
+            ret = sense_cat;
+            break;
+        case SG_LIB_CAT_RECOVERED:
+        case SG_LIB_CAT_NO_SENSE:
+            ret = 0;
+            break;
+        default:
+            ret = -1;
+            break;
+        }
+    } else {
+        if ((verbose > 2) && (ret > 0)) {
+            fprintf(sg_warnings_strm, "    read block limits: response%s\n",
+                    (ret > 256 ? ", first 256 bytes" : ""));
+            dStrHex((const char *)resp, (ret > 256 ? 256 : ret), -1);
+        }
+        ret = 0;
+    }
+    destruct_scsi_pt_obj(ptvp);
+    return ret;
+}
